/*   
 * Copyright (C) 2006 Daniel Fernandez <ergot86@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License version 2 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */
#include <signal.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/ptrace.h>
#include <sys/stat.h>
#include <linux/user.h>
#include <linux/stddef.h>
#include <linux/ptrace.h>
#include <linux/limits.h>
#include <unistd.h>
#include <fcntl.h>

#define FALSE 0
#define TRUE  !FALSE

#define PTRACE_O_ALL PTRACE_O_TRACESYSGOOD | PTRACE_O_TRACEFORK | PTRACE_O_TRACEVFORK | PTRACE_O_TRACECLONE | \
        PTRACE_O_TRACEEXEC | PTRACE_O_TRACEVFORKDONE | PTRACE_O_TRACEEXIT

/* quedan por aca... */
#define PTRACE_MPROTECT         0x4204
#define PTRACE_GETPAGEPROT      0x4205
/* ...hasta que encuentren un hogar */
 
#define VM_READ   1
#define VM_WRITE  2
#define VM_EXEC   4
#define VM_SHARED 8

#ifdef __i386__
#define REG_IP dbg->regs.eip
#define REG_SP dbg->regs.esp
#endif

#ifdef __x86_64__
#define REG_IP dbg->regs.rip
#define REG_SP dbg->regs.rsp

/* Hack: en x86_64 la estructura user_i387_struct incluye las extensiones */
#define user_fxsr_struct user_i387_struct
#undef PTRACE_GETFPXREGS
#undef PTRACE_SETFPXREGS
#define PTRACE_GETFPXREGS PTRACE_GETFPREGS
#define PTRACE_SETFPXREGS PTRACE_SETFPREGS

#define REX     '\x40'
#define REXB    '\x41'
#define REXX    '\x42'
#define REXXB   '\x43'
#define REXR    '\x44'
#define REXRB   '\x45'
#define REXRX   '\x46'
#define REXRXB  '\x47'
#define REXW    '\x48'
#define REXWB   '\x49'
#define REXWX   '\x4a'
#define REXWXB  '\x4b'
#define REXWR   '\x4c'
#define REXWRB  '\x4d'
#define REXWRX  '\x4e'
#define REXWRXB '\x4f'

#endif

#define PREFIX_CS '\x2e'
#define PREFIX_DS '\x3e'
#define PREFIX_ES '\x26'
#define PREFIX_FS '\x64'
#define PREFIX_GS '\x65'
#define PREFIX_SS '\x36'

#define LOCK  '\xf0'
#define REPZ  '\xf3'
#define REPNZ '\xf2'
#define ADDRS '\x67'
#define OPERS '\x66'

#define CALL_NI '\xe8'
#define CALL_FD '\x9a'
#define CALL    '\xff'

#define INT_3 '\xcc'

#define WP_READ   1
#define WP_WRITE  2
#define WP_EXEC   4
#define WP_ACCESS 8

struct ptrace_mprotect_args
{
  unsigned long addr ;
  size_t len ;
  unsigned long prot ;
} ;

struct vm_map
{
  unsigned long start ;
  unsigned long end ;
  unsigned long flags ;
  unsigned long offset ;
  dev_t dev ;
  ino_t inode ;
  char file[PATH_MAX] ;
} ;

struct BreakPoint
{
  struct BreakPoint *next ;
  unsigned long addr ;
  char op ;
  char bp ; /* guardamos el opcode del bp por si es cambiado con SetBpOpcode() */
} ;

struct WatchPoint
{
  struct WatchPoint *next ;
  unsigned long addr ;
  unsigned long prot ;    /* flags originales */
  char wpa[PAGE_SIZE/2] ; /* nibble con WP flags por byte de la pagina,
                             prefiero algo de complejidad extra en el codigo
                             a cambio de ahorro de memoria */
} ;

struct Debug
{
  pid_t pid ;
  siginfo_t sig ;
  struct user_regs_struct regs ;
  struct user_fxsr_struct fregs ;
  struct BreakPoint *bps ;
  struct WatchPoint *wps ;
  int u_debugreg[8] ;
  int mem ;                          /* descriptor para ReadMem */
  int sysemu ;                       /* evita syscalls si el flag esta activado */
  void (*DbgH) (struct Debug *dbg) ; /* Manejador customizable */
} ;

/* usada para wpa */
static inline int IsZero(long *addr, size_t size)
{
  long *ad = addr+size ;
  long tmp = 0 ;

  do {
    tmp |= addr[0] | addr[1] | addr[2] | addr[3] ;
  } while ((addr += 4) < ad) ;

  return (tmp)? TRUE:FALSE ;
}

static inline int IsSyscall(struct Debug *dbg)
{
  return (dbg->sig.si_code == (SIGTRAP|0x80))? TRUE:FALSE ;
}

static inline int IsEvent(struct Debug *dbg, const int event)
{
  return (dbg->sig.si_code == ((event << 8)|SIGTRAP))? TRUE:FALSE ;
}

static inline int IsTrap(struct Debug *dbg, const int event)
{
  return (dbg->sig.si_code == SI_KERNEL && dbg->sig.si_signo == SIGTRAP)? TRUE:FALSE ;
}

/*
 ejemplo:
 
 if (IsEvent(dbg,PTRACE_EVENT_FORK) | IsEvent(dbg,PTRACE_EVENT_VFORK) | IsEvent(dbg,PTRACE_EVENT_CLONE))
  GetChildPid(dbg,&new_pid) ;

*/

static inline void GetChildPid(struct Debug *dbg,pid_t *pid)
{
  ptrace(PTRACE_GETEVENTMSG,dbg->pid,NULL,&pid) ;
}

/*
 ejemplo:

  if (IsEvent(dbg,PTRACE_EVENT_EXIT) == TRUE)
    GetExitCode(dbg.&code) ;

*/

static inline void GetExitCode(struct Debug *dbg,int *code)
{
  ptrace(PTRACE_GETEVENTMSG,dbg->pid,NULL,&code) ;
}


/* mem.c */
int WriteMem(unsigned long addr, size_t sz, unsigned char *buff, struct Debug *dbg) ;
int ReadMem(unsigned long addr, size_t sz, char *buff, struct Debug *dbg) ;
struct WatchPoint * FindWp(unsigned long addr,struct Debug *dbg) ;
int StillWpSet(struct WatchPoint *wps,struct Debug *dbg) ;
int IsWp(struct Debug *dbg) ;
int AddWp(struct WatchPoint *wps,unsigned long addr,size_t size,unsigned long flags,struct Debug *dbg) ;
int SetWp(unsigned long addr,size_t size,unsigned long flags,struct Debug *dbg) ;
int RemoveWp(unsigned long addr,size_t size,struct Debug *dbg) ;
unsigned int RemoveAllWp(struct Debug *dbg) ;
int ResetWp(unsigned long addr,struct Debug *dbg) ;
unsigned int ResetAllWp(struct Debug *dbg) ;

/* trace.c */
int StepIn(struct Debug *dbg) ;
int StepOver(struct Debug *dbg) ;
int StepFromWp(struct Debug *dbg) ;
int Run(struct Debug *dbg) ;
int Syscall(struct Debug *dbg) ;
char SetBpOpcode(char op) ;
struct BreakPoint * FindBp(unsigned long addr,struct Debug *dbg) ;
int StillBpSet(struct BreakPoint *bps,struct Debug *dbg) ;
int IsBp(struct Debug *dbg) ;
int SetBp(unsigned long addr,struct Debug *dbg) ;
int RemoveBp(unsigned long addr,struct Debug *dbg) ;
unsigned int RemoveAllBp(struct Debug *dbg) ;
int ResetBp(unsigned long addr,struct Debug *dbg) ;
unsigned int ResetAllBp(struct Debug *dbg) ;

/* dbg.c */
void DSnap(struct Debug *dbg) ;
void SetRegs(struct Debug *dbg) ;
int Trace(struct Debug *dbg, unsigned long fail_action) ;
int TraceFile(const char *file, char *argv[], char *env[], struct Debug *dbg) ;
int Attach(struct Debug *dbg) ;
int Detach(struct Debug *dbg) ;
int Kill(struct Debug *dbg) ;
int SendSignal(struct Debug *dbg) ;
int GetMapInfo(unsigned long addr,struct vm_map *map,struct Debug *dbg) ;

